#region License
// Product:
// MCM (Message-Command-Message)
// mcm-framework.net
// 
// Author:
// Pawel Idzikowski
// Cologne, Germany
// idzikowski@polenter.com
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the author(s) be held liable for any damages arising from
// the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//       claim that you wrote the original software. If you use this software
//       in a product, an acknowledgment in the product documentation would be
//       appreciated but is not required.
// 
//     2. Altered source versions must be plainly marked as such, and must not
//       be misrepresented as being the original software.
// 
//     3. This notice may not be removed or altered from any source distribution.
// 
#endregion
using System;

namespace Polenter.Mcm
{
    /// <summary>
    /// Each command contains Argument, Result and Error properties.
    /// Use CommandAsyncExecutor to execute a Command asynchronously.
    /// The command should not access external components and services directly.
    /// The whole environment the command uses should be defined in the
    /// Argument property. 
    /// Such isolation makes unit tests possible. 
    /// </summary>
    /// <typeparam name="TArgument"></typeparam>
    /// <typeparam name="TResult"></typeparam>
    public abstract class Command<TArgument, TResult> : IMcmCommand
    {
        /// <summary>
        /// Standard constructor
        /// </summary>
        protected Command()
        {
        }

        /// <summary>
        /// Constructor with an argument
        /// </summary>
        /// <param name="argument"></param>
        protected Command(TArgument argument)
        {
            Argument = argument;
        }

        ///<summary>
        /// Any object can be the argument
        ///</summary>
        public TArgument Argument { get; private set; }

        ///<summary>
        /// Contains result after executing
        ///</summary>
        public TResult Result { get; protected set; }

        #region IMcmCommand Members

        ///<summary>
        /// Contains last error occured by the executing
        ///</summary>
        public Exception Error { get; private set; }

        ///<summary>
        /// Tells if last executing has produced an error
        ///</summary>
        public bool HasError
        {
            get { return Error != null; }
        }

        /// <summary>
        /// Start the command. Error and result are reset.
        /// </summary>
        /// <returns>false if there is an error occured</returns>
        public bool Execute()
        {
            Error = null;
            Result = default(TResult);
            try
            {
                ExecuteCore();
                return true;
            }
            catch (Exception ex)
            {
                Error = ex;
                return false;
            }
            finally
            {
                invokeExecuteCompleted(EventArgs.Empty);
            }
        }

        ///<summary>
        /// Invoked after execution. Result contains the working result, Error - the last error if any.
        ///</summary>
        public event EventHandler ExecuteCompleted;

        #endregion

        /// <summary>
        /// try..catch is already implemented in the base class
        /// </summary>
        protected abstract void ExecuteCore();

        private void invokeExecuteCompleted(EventArgs e)
        {
            EventHandler handler = ExecuteCompleted;
            if (handler != null) handler(this, e);
        }
    }
}